--[[
TODO
	- resizable frame?
	- convert currency link to id for save? conflict with item ids?
	- auto mark items as soulbound when bop or soulbound?
	- xmog bank info?
	
	-options menu
		--list of characters and way to clear data (renames, xfers, etc)
			-use gold list
		--sorting options?
]]

--make a copy of a table
local function copyTable(oldTable, meta, deep)
	local result = {}

	for key, value in pairs(oldTable) do
		if type(value) == "table" and deep then
			result[key] = copyTable(value, meta, deep)
		else
			result[key] = value
		end
	end

    -- copy the metatable, if there is one
	if meta then
		setmetatable(result, getmetatable(oldTable))
	end
	
	return result
end

--merge 2 tables. no protection of old tables values implemented
local function mergeTable(oldTable, newTable)
	for key, value in pairs(newTable) do
		oldTable[key] = value
	end
end

local ReagentTracker = addon

--main frame
local addon = CreateFrame("FRAME", "ReagentTracker", UIParent)
addon:RegisterEvent("ADDON_LOADED")

addon:SetSize(360, 180)
addon:SetPoint("CENTER")
addon:SetFrameStrata("HIGH")

addon:SetMovable(true)
addon:EnableMouse(true)
addon:RegisterForDrag("LeftButton")

addon:SetScript("OnEvent", function(self, event, ...) self:OnEvent(event, ...) end)

--event handling
function addon:OnEvent(event, ...)
	if event == "ADDON_LOADED" then
		local name = ...
		if name == "ReagentTracker" then
			self:UnregisterEvent("ADDON_LOADED")
			
			REAGENT_STATS = REAGENT_STATS or {}
			self.reagents = REAGENT_STATS
			
			--load or defaults
			local defaultSettings = {
				showTooltips = false,
				showQuantityInfo = true,
				showHistoryInfo = true,
				showUntrackedQuantity = false,
				
				watchList = {},
				currencyWatchList = { gold = {} },
			}
			REAGENT_SETTINGS = REAGENT_SETTINGS or copyTable(defaultSettings, true, true)
			self.settings = REAGENT_SETTINGS
			
			--initialize new settings
			for setting, value in pairs(defaultSettings) do
				if self.settings[setting] == nil then
					self.settings[setting] = value
				end
			end
						
			--apply sorting system if it isnt
			if not self.settings.watchListSorted then
				self:InitializeListSort()			
			end
			
			--remove link and texture from saved data
			if not self.settings.textureCleared then
				self:RemoveSavedTextures()
			end
						
			--list of color code prefixes for formatting purposes
			self.prefix = {
				reg = {
					less = "|cFFFF0000",
					more = "|cFF00FF00+",
					equal = "|cFF0099FF"
				},
				gold = {
					less = "|cFFFF0000",
					more = "|cFF00FF00",
					equal = "|cFF0099FF"
				},
				trend = {
					less = "|cFFFF0000",
					more = "|cFF0099FF",
					equal = "|cFF0099FF"
				}
			}
			
			--useful player info
			self.player = {
				name = UnitName("player"),
				faction = UnitFactionGroup("player"),
				server = GetRealmName(),
			}
			
			--add server and faction info to reagents.  old data is updated to first server and faction you join
			if not self.settings.formatUpdate then
				self:UpdateReagentFormat()
			end
			
			self:InitializeServerFactionInfo()
			
			--add gold to currency tab
			if not self.settings.goldAdded then
				self.settings.goldAdded = true
				self.settings.currencyWatchList["gold"] = {}
				if not tContains(self.settings.sortedCurrencyList, "gold") then
					table.insert(self.settings.sortedCurrencyList, "gold")
				end
			end
		
			self:RegisterEvent("PLAYER_ENTERING_WORLD")
		end
		
	elseif event == "PLAYER_ENTERING_WORLD" then
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		
		--convert old date system
		if not self.settings.newDateSystem then
			self:ConvertDateStamps()
		end
		
		--convert old table to single value
		if not self.settings.newCountSystem then
			self:ConvertItemCounts()
		end
			
		self:LoadTrackerFrame()
		self:LoadEditPopup()
		self:LoadCurrencyTabChanges()
		self:LoadReagentTooltip()
		self:HookScripts()
		self:LoadTimers()
				
		PanelTemplates_SetNumTabs(self, 3)
		PanelTemplates_SetTab(self, 1)
		self.tabDisplay = 1
		
		self.visibleWatchList = self.settings.watchList
		self.visibleSortedList = self.settings.sortedReagentList
		
		self.affectedMailItems = {}
				
		--watch for reagent updates
		self:RegisterEvent("BAG_UPDATE")										--char bags and bank bags
		self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")					--main bank slots
		self:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")		--reagent bank
		
		self:RegisterEvent("PLAYER_MONEY")									--gold
		self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")					--other currencies
		
		self:RegisterEvent("MAIL_SEND_INFO_UPDATE")						--build list of any tracked items that were placed into mail
		self:RegisterEvent("MAIL_CLOSED")										--update the above list when the mail window closes. this ensures that the items are no longer part of your inventory still when they are counted again.
			
		self:SetScript("OnShow", function() 
			self:UpdateTrackerFrame()
		end)
		self:SetupChatCommands()
		
		--login reagent check and purge
		self:ReagentRollCall()
		self:DeleteOldDateData()
		
	elseif event == "BAG_UPDATE" then
		self:AddBagIDCheck(...)

	elseif event == "PLAYERBANKSLOTS_CHANGED" then
		local slotID = ...
		self:AddBagIDCheck(BANK_CONTAINER, slotID)
	
	elseif event == "PLAYERREAGENTBANKSLOTS_CHANGED" then
		local slotID = ...
		self:AddBagIDCheck(REAGENTBANK_CONTAINER, slotID)
	
	elseif event == "PLAYER_MONEY" or event == "CURRENCY_DISPLAY_UPDATE" then
		self.currencyUpdate = true
	
	elseif event =="MAIL_SEND_INFO_UPDATE" then
		self:BuildMailItemList(self.affectedMailItems)
	
	elseif event == "MAIL_CLOSED" then
		self:ReagentUpdate(self.affectedMailItems, true)
		self.affectedMailItems = {}
		
	end
end

--grab latest reagent info for the character and update display
function addon:ReagentUpdate(updateList, ignoreCurrency)
	self:ReagentRollCall(updateList, ignoreCurrency)
	if self.TrackerFrame:IsVisible() then
		self:UpdateTrackerFrame()
	end
end

--setup chat commands
function addon:SetupChatCommands()
	--Setup slash commands for chatbox
	SLASH_REAGENTTRACKER1 = "/rt"
	SLASH_REAGENTTRACKER2 = "/reagenttracker"
	SLASH_REAGENTTRACKER3 = "/reagent"
	function SlashCmdList.REAGENTTRACKER(msg) self:ChatCommands(msg) end
end

function addon:ChatCommands(msg)
	--msg = strlower(msg)
	local msg, msgInfo = strsplit(" ", msg, 2)
	if msg == "add" then
		if msgInfo then
			self:AddReagent(msgInfo)
		end
	elseif msg == "remove" then
		self:RemoveReagent(msgInfo)
	
	elseif msg == "update" then
		self:UpdateReagentName(msgInfo)
		
	else
		addon:ToggleVisible()
	end
end

--REAGENTS
--add reagent to watch list
function addon:AddReagent(reagent)
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(reagent)
	local itemID
	if link then
		itemID = tonumber(link:match("|Hitem:(%d+)"))
	end
	local reagentName = itemID or name or reagent
	
	if reagentName then
		if not self.settings.watchList[reagentName] then
			self.settings.watchList[reagentName] = {}
			self.settings.watchList[reagentName].itemLink = link
			self.settings.watchList[reagentName].itemID = itemID
			table.insert(self.settings.sortedReagentList, reagentName)
		end
	
		print((link or reagentName).." has been added to the watch list.")
		self:ReagentUpdate({[reagentName] = true}, true)
	end
end

--remove reagent from watch list
function addon:RemoveReagent(reagent)
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(reagent)
	local itemID
	if link then
		itemID = tonumber(link:match("|Hitem:(%d+)"))
	end	
	local reagentName = itemID or name or reagent
	
	if self.settings.watchList[reagentName] then
		self.settings.watchList[reagentName] = nil
		tDeleteItem(self.settings.sortedReagentList, reagentName)
		
		print((link or reagentName).." has been removed from the watch list.")
		self:ReagentUpdate({}, true)
	else
		print(reagent.." could not be found.")
	end
end

--update a link or text name into an itemID
function addon:UpdateReagentName(reagent)
	if not tonumber(reagent) then
		local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(reagent)
		local itemID
		if link then
			itemID = tonumber(link:match("|Hitem:(%d+)"))
		end
		local reagentName = name or reagent
		
		if itemID then
			--update item watchlist
			if self.settings.watchList[reagentName] then
				self.settings.watchList[itemID] = copyTable(self.settings.watchList[reagentName], true, true)
				self.settings.watchList[reagentName] = nil
			end
			
			--update actual character data location
			if self.localReagents[reagentName] then
				self.localReagents[itemID] = copyTable(self.localReagents[reagentName], true, true)
				self.localReagents[reagentName] = nil
			end
			
			--update sorted list
			if tContains(self.settings.sortedReagentList, reagentName) then
				local function findReagentID(reagent)
					local sortedID
					for id, info in pairs(self.settings.sortedReagentList) do
						if info == reagent then
							return id
						end
					end
				end

				table.insert(self.settings.sortedReagentList, findReagentID(reagentName), itemID)
				tDeleteItem(self.settings.sortedReagentList, reagentName)
			end
			
			return itemID
		end
	end
end

--CURRENCIES
--toggle currency when button clicked
function addon:ToggleCurrencyTracking(currencyID)
	if self.ReagentTrackerCurrencyButton:GetChecked() then
		self:AddCurrency(currencyID)
		PlaySound("igMainMenuOptionCheckBoxOn")
	else
		self:RemoveCurrency(currencyID)
		PlaySound("igMainMenuOptionCheckBoxOff")
	end
	
	--update the currency list if its visible
	if self:IsVisible() and self.tabDisplay == 2 then
		self:Hide()
		self:Show()
	end
end

--add a new currency
function addon:AddCurrency(currencyID)
	local itemLink = GetCurrencyListLink(currencyID)
	local name, currentAmount, texture, earnedThisWeek, weeklyMax, totalMax, isDiscovered, rarity = GetCurrencyInfo(itemLink)
	
	if itemLink then
		if not self.settings.currencyWatchList[itemLink] then
			self.settings.currencyWatchList[itemLink] = {}
			table.insert(self.settings.sortedCurrencyList, itemLink)
			self:ReagentUpdate({}, false)
		end
	end
end

--remove a tracked currency
function addon:RemoveCurrency(currencyID)
	--edit menu popup
	if self.settings.currencyWatchList[currencyID] then
		self.settings.currencyWatchList[currencyID] = nil
		tDeleteItem(self.settings.sortedCurrencyList, currencyID)
		
		--update frame
		self:Hide()
		self:Show()
		
		return true
	end
	
	--toggle checkbox in currency list
	local itemLink = GetCurrencyListLink(currencyID)
	if self.settings.currencyWatchList[itemLink] then
		self.settings.currencyWatchList[itemLink] = nil
		tDeleteItem(self.settings.sortedCurrencyList, itemLink)
		self:ReagentUpdate({}, false)
	end
end

--update watchlist with reagent info for display purposes
function addon:SaveReagentLinkInfo(reagent)
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(reagent)
	
	if link then
		self.settings.watchList[reagent].itemLink = link
	end
	return link or self.settings.watchList[reagent].itemLink
end

--hook functions and stuff
function addon:HookScripts()
	if not self.hooked then
		
		--update my currency toggle button when the other token popup buttons are
		hooksecurefunc("TokenFrame_UpdatePopup", function() self:UpdateCurrencyButton() end)
		
		--hook tooltip show and hide
		GameTooltip:HookScript("OnShow", function(frame) self:ShowReagentTooltip(frame) end)
		GameTooltip:HookScript("OnHide", function() self.ReagentTooltip:Hide() end)
		
		ItemRefTooltip:HookScript("OnShow", function(frame) self:ShowReagentTooltip(frame) end)
		ItemRefTooltip:HookScript("OnHide", function() self.ReagentTooltip2:Hide() end)
		
		--intercept logout for currency updates. items were handled by events. check mail list incase they somehow logged out without closing the window
		local function logoutChecks()
			if self.currencyUpdate or self.currencyUpdateDelayed then
				self:ReagentRollCall({}, false)
			end
			self:ReagentRollCall(self.affectedMailItems, true)
		end
		
		local Orig_Logout = Logout
		function Logout()
			logoutChecks()
			Orig_Logout()
		end
		
		local Orig_ForceQuit = ForceQuit
		function ForceQuit()
			logoutChecks()
			Orig_ForceQuit()
		end
		
		local Orig_Quit = Quit
		function Quit()			
			logoutChecks()
			Orig_Quit()
		end
		
		self.hooked = true
	end
end

--toggle visibility of the main frame
function addon:ToggleVisible(hide)
	if self:IsVisible() or hide then
		PlaySound("igCharacterInfoClose")
		self:Hide()
	else
		PlaySound("igCharacterInfoOpen")
		self:Show()
	end
end

--[[collect data on watched reagents and save for current character
		updateList - list containing reagents to update.  nil defaults to the entire list. empty table ignores reagents
		ignoreCurrency - flag to update currency list or not
		quitFunc	- function called on logout after update is done
		
	called on login, logout, bag update or currency update
		login initializes reagents best and catches errors from bad quits
		bag and currency updates happen as they are used/gained
		logout catches delayed currency updates and any mail items left over
]]
function addon:ReagentRollCall(updateList, ignoreCurrency)
	local dateStamp = self:GetDateStamp()
	local playerName = UnitName("player")
	
	updateList = updateList or self.settings.watchList
	
	--reagents
	for reagent in pairs(updateList) do
		self:InitializeReagent(reagent)
		local info = self.localReagents[reagent]
		local count
		if reagent == "gold" then
			count = GetMoney()
		else
			count = GetItemCount(info.itemLink or reagent, true, true)
		end
		
		info[playerName].currentCount = count
		info[playerName].data[dateStamp] = count
	end
	
	--currencies
	if not ignoreCurrency then
		for currency in pairs(self.settings.currencyWatchList) do
				self:InitializeReagent(currency)
				local info = self.localReagents[currency]
				local _, count
			if currency == "gold" then
				count = GetMoney()
			else
				_, count = GetCurrencyInfo(currency)
			end
				info[playerName].currentCount = count
				info[playerName].data[dateStamp] = count
		end
	end
end

--initialize tables and values for a reagent
function addon:InitializeReagent(reagent)
	local playerName = UnitName("player")
	
	self.localReagents[reagent] = self.localReagents[reagent] or {}
	
	if not self.localReagents[reagent][playerName] then
		self.localReagents[reagent][playerName] = {}							--table for char
		self.localReagents[reagent][playerName].currentCount = 0			--latest total count
		self.localReagents[reagent][playerName].data = {}					--table to contain dates and counts for trending
		
		return true
	end
	
	return false
end

--create sorted list of watched items for display purposes
function addon:InitializeListSort()
	--sort reagents
	local sortedTable = {}
	for reagent in pairs(self.settings.watchList) do
		table.insert(sortedTable, reagent)
	end
	table.sort(sortedTable)
	self.settings.sortedReagentList = sortedTable
	
	--sort currencies
	sortedTable = {}
	for currency in pairs(self.settings.currencyWatchList) do
		table.insert(sortedTable, currency)
	end
	table.sort(sortedTable)
	self.settings.sortedCurrencyList = sortedTable

	self.settings.watchListSorted = true
end

--return a table containing the dateStrings of a time period, oldest first
function addon:PreviousTimePeriod(totalOffset, timeStamp)
	totalOffset = totalOffset or -13							--default to 2 weeks
	timeStamp = timeStamp or self:GetDateStamp()		--today's date
	
	local timePeriod = {}
	for offset = totalOffset, 0, 1 do
		local offsetDate = self:GetDateStamp(timeStamp, offset)
		table.insert(timePeriod, offsetDate)
	end
	
	return timePeriod
end

--remove any data for dates older than 2 weeks
function addon:DeleteOldDateData()
	local immuneTimePeriod = self:PreviousTimePeriod()
	
	for reagent, characterInfo in pairs(self.localReagents) do
		for character, info in pairs(characterInfo) do
			if type(info) == "table" then
				for dateString, itemCount in pairs(info.data) do
					if not tContains(immuneTimePeriod, dateString) then
						info.data[dateString] = nil
					end
				end
			end
		end
	end
end

--return a character reagent count for a specific date. search back then forward in time for a value before giving up
function addon:CharacterReagentCountSearch(reagent, character, dateString, search)
	--default to current character
	character = character or UnitName("player")
	
	--default to today
	dateString = dateString or self:GetDateStamp()
	
	--look to past and future days for values or return todays only
	if search then
		local count = false
		local offset = 0
		repeat
			--look into past for last known amount
			local offsetDate = self:GetDateStamp(dateString, offset)
			count = self:CharacterReagentCountSearch(reagent, character, offsetDate, false)
			offset = offset - 1
		until count or offset < -13		--value found or 2 weeks from date
		
		if count then
			return count
		else
			--no known past amount found, try future
			local offset = 1
			repeat
				local offsetDate = self:GetDateStamp(dateString, offset)
				count = self:CharacterReagentCountSearch(reagent, character, offsetDate, false)
				offset = offset + 1
			until count or offset > 13		--value found or 2 weeks from date
		end
		
		return count or 0
	else
		--return specific date only
		if self.localReagents[reagent][character].data[dateString] then
			return self.localReagents[reagent][character].data[dateString]
		else
			return false
		end
	end
end

--return the sum of all character reagent counts for a specific date using search func
function addon:AllCharReagentCountSearch(reagent, dateString, soulbound)
	local totalCount = 0
	local individualCounts = {}
	if soulbound then
		--current character only
		local character = UnitName("player")
		totalCount = self:CharacterReagentCountSearch(reagent, character, dateString, true) or 0
		individualCounts[character] = totalCount
	else
		--everyone
		for character, info in pairs(self.localReagents[reagent]) do
			if type(info) == "table" then
				local charCount = self:CharacterReagentCountSearch(reagent, character, dateString, true) or 0
				
				totalCount = totalCount + charCount
				individualCounts[character] = charCount
			end
		end
	end
	
	return totalCount, individualCounts
end

--return the sum of all current character reagent counts using .currentCount
function addon:AllCharReagentCountCurrent(reagent)
	local totalCount = 0
	local individualCounts = {}

	for character, info in pairs(self.localReagents[reagent]) do
		if type(info) == "table" then
			local charCount = self.localReagents[reagent][character].currentCount or 0
			
			totalCount = totalCount + charCount
			individualCounts[character] = charCount
		end
	end
	
	return totalCount or 0, individualCounts
end

--return the slope/trend and history for a certain reagent
function addon:GetTimePeriodInfo(reagent, soulbound)
	local weekDays = self:PreviousTimePeriod()		--defaults to a period 2 weeks back from today
	local dailyTotals = {}
	for _, dateString in pairs(weekDays) do
		local totalCount = self:AllCharReagentCountSearch(reagent, dateString, soulbound)
		table.insert(dailyTotals, totalCount)
	end
	local sumX, sumY, sumXY, sumsqX = 0, 0, 0, 0
	local num = #dailyTotals
	for x, y in pairs(dailyTotals) do
		sumXY = sumXY + (x * y)
		sumX = sumX + x
		sumY = sumY + y
		sumsqX = sumsqX + (x ^ 2)
	end
	
	local slope = ( (num * sumXY) - (sumX * sumY) ) / ( (num * sumsqX) - (sumX ^ 2) )
	
	local info = {
		trend = slope,
		history = dailyTotals,
		dates = weekDays,
	}
	return info
end

--build scroll frame to house notes
function addon:LoadTrackerFrame()
	local frame, fontString, button, texture, dropDown
	
	self:SetBackdrop({
		bgFile="Interface\\CHATFRAME\\CHATFRAMEBACKGROUND", 
		edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
		tile=1, tileSize=10, edgeSize=32, 
		insets={left=10, right=10, top=10, bottom=10}
		})
	self:SetBackdropColor(0, 0, 0, 0.6)

	self:SetScript("OnDragStart", function() self:StartMoving() end)
	self:SetScript("OnDragStop", function() self:StopMovingOrSizing() end)
	
	--title frame
	frame = CreateFrame("FRAME", "ReagentTracker_Title", self)
	self.ReagentTrackerTitleFrame = frame
	frame:SetPoint("TOPLEFT", self,"TOPLEFT", 16, -14)
	frame:SetPoint("TOPRIGHT", self,"TOPRIGHT", -16 - 20, -14)
	frame:SetHeight(20)
	
	--reagent title
	fontString = frame:CreateFontString()
	self.ReagentTrackerReagentTitle = fontString
	fontString:SetPoint("LEFT", self.ReagentTrackerTitleFrame, "LEFT", 4, 0)
	fontString:SetFontObject("GameFontNormal")
	fontString:SetText("Reagent")
	
	--trend title
	fontString = frame:CreateFontString()
	self.ReagentTrackerTrendTitle = fontString
	fontString:SetPoint("RIGHT", self.ReagentTrackerTitleFrame, "RIGHT", -4, 0)
	fontString:SetFontObject("GameFontNormal")
	fontString:SetWidth(50)
	fontString:SetJustifyH("RIGHT")
	fontString:SetText("Trend")
	
	--quantity title
	fontString = frame:CreateFontString()
	self.ReagentTrackerQuantityTitle = fontString
	fontString:SetPoint("RIGHT", self.ReagentTrackerTrendTitle, "LEFT", -4, 0)
	fontString:SetFontObject("GameFontNormal")
	fontString:SetWidth(60)
	fontString:SetJustifyH("CENTER")
	fontString:SetText("Quantity")
		
	--close button
	button = CreateFrame("BUTTON", "ReagentTracker_CloseButton", self, "UIPanelCloseButton")
	button:SetPoint("TOPRIGHT", -3, -3)
	button:SetScript("OnClick", function() self:ToggleVisible(true) end)
	
	texture = self:CreateTexture()
	texture:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Corner")
	texture:SetPoint("TOPRIGHT", -6, -7)
	texture:SetSize(32, 32)
	texture:SetDrawLayer("OVERLAY")
	
	--drag frame
	frame = CreateFrame("FRAME", "ReagentTracker_DragFrame", self)
	self.DragFrame = frame
	frame:SetSize(3, 3)
	frame:Hide()
	frame:SetScript("OnUpdate", function(frame) self:UpdateDragFrame(frame) end)
	frame:SetScript("OnHide", function() self.DragFrameTooltip:Hide() ResetCursor() end)
	
	--tracker frame
	frame = CreateFrame("FRAME", "ReagentTracker_Main", self)
	self.TrackerFrame = frame
	frame:SetPoint("TOPLEFT", self,"TOPLEFT", 16, -14 - 20)
	frame:SetPoint("BOTTOMRIGHT", self,"BOTTOMRIGHT", -36, 12)
	frame:SetScript("OnShow", function()
		self.OptionsFrame:Hide()
		
		--currency check was flagged
		if self.currencyUpdateDelayed then
			self.currencyUpdateDelayed = false
			self:ReagentRollCall({}, false)
			if self.tabDisplay == 2 then
				self:UpdateTrackerFrame()
			end
		end
	end)
	
	--scroll frame
	frame = CreateFrame("SCROLLFRAME", "ReagentTracker_Window", self.TrackerFrame, "FauxScrollFrameTemplate")
	self.ReagentTrackerWindow = frame
	frame:SetAllPoints(self.TrackerFrame)
	frame:SetScript("OnVerticalScroll", function(frame, offset)
		FauxScrollFrame_OnVerticalScroll(frame, offset, 20, function() self:UpdateTrackerFrame() end)
		if self.tabDisplay == 1 then
			self.prevReagentOffset = offset
		elseif self.tabDisplay == 2 then
			self.prevCurrencyOffset = offset
		end
	end)
	frame:EnableMouseWheel(true)
	
	--Make the content frame
	frame = CreateFrame("FRAME", "ReagentTracker_Content", self.TrackerFrame)
	self.ReagentTrackerContent = frame
	frame:SetPoint("TOPLEFT", self.ReagentTrackerWindow)
	frame:SetWidth(self.ReagentTrackerWindow:GetWidth())
	frame:SetHeight(self.ReagentTrackerWindow:GetHeight())
	frame:SetScript("OnMouseUp", function() self:GetCursorReagent() end)
	frame:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
	
	--make container frames
	self.ReagentDisplay = {}
	for id = 1, 6 do
		self:UpdateTrackerFrameCreateFrame(id)
	end
		
	--options frame
	frame = CreateFrame("FRAME", "ReagentTracker_Options", self)
	self.OptionsFrame = frame
	frame:SetPoint("TOPLEFT", self,"TOPLEFT", 16, -14)
	frame:SetPoint("BOTTOMRIGHT", self,"BOTTOMRIGHT", -36, 12)
	frame:SetScript("OnShow", function() 
		self.TrackerFrame:Hide()
	end)
	frame:Hide()
	
	--reagent addition button
	button = CreateFrame("BUTTON", "ReagentTracker_AddButton", self.OptionsFrame, "SimplePopupButtonTemplate")
	self.AddReagentButton = button
	button:SetPoint("TOPLEFT", 5, -5 - 20)
	button:SetSize(36, 36)
	button:SetHighlightTexture("Interface\\BUTTONS\\ButtonHilight-Square")
	button:SetScript("OnClick", function() self:GetCursorReagent() end)
	button:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Drag items here to add them to the tracker.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	--dropdown containing known character list
	dropDown = CreateFrame("BUTTON", "ReagentTrackerOptions_CharDropdown", self.OptionsFrame, "Lib_UIDropDownMenuTemplate")
	self.CharDropdown = dropDown
	dropDown:SetPoint("BOTTOMLEFT", self.OptionsFrame, -14, 0)
	Lib_UIDropDownMenu_Initialize(dropDown, function(dropDown, level) self:CharacterListDropdownInitialization(dropDown, level) end)
	Lib_UIDropDownMenu_SetWidth(dropDown, 100)
	Lib_UIDropDownMenu_SetButtonWidth(dropDown, 100)
	Lib_UIDropDownMenu_JustifyText(dropDown, "LEFT")
	dropDown.Disable = function (self) Lib_UIDropDownMenu_DisableDropDown(self) end
	dropDown.Enable = function (self) Lib_UIDropDownMenu_EnableDropDown(self) end
	Lib_UIDropDownMenu_SetText(dropDown, "Characters")
	dropDown:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Select a character to erase data.") end)
	dropDown:SetScript("OnLeave", function() GameTooltip:Hide() end)
		
	--character data erasing button
	button = CreateFrame("BUTTON", "ReagentTrackerOptions_DeleteCharacterButton", self.OptionsFrame, "UIPanelButtonTemplate")
	self.DeleteCharButton = button
	button:SetPoint("LEFT", self.CharDropdown, "RIGHT", -5, 3)
	button:SetSize(80, 20)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Erase Data")
	button:SetScript("OnClick", function()
		StaticPopupDialogs["ReagentTrackerConfirmCharacterDelete"] = {	
			text = string.format("Erasing data for |cFFFF0000%s|r is irreversible.  Are you sure you want to do this?", Lib_UIDropDownMenu_GetSelectedValue(self.CharDropdown) or "Unknown"),
			button1 = YES,
			button2 = NO,
			OnAccept = function() 
				self:DeleteCharacterData(Lib_UIDropDownMenu_GetSelectedValue(self.CharDropdown))
			end,
			OnCancel = function()
				--reset 
				Lib_UIDropDownMenu_SetSelectedValue(self.CharDropdown, nil)
				Lib_UIDropDownMenu_SetText(self.CharDropdown, "Characters")
				self.DeleteCharButton:Disable()
			end,
			hideOnEscape = 1,
			timeout = 0,
		}
		StaticPopup_Show("ReagentTrackerConfirmCharacterDelete")
	end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, string.format("Erase data for |cFFFF0000%s|r.  A popup will appear to confirm.", Lib_UIDropDownMenu_GetSelectedValue(self.CharDropdown) or "Unknown")) end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	button:Disable()

	--show tooltip button
	--child button control function
	local function updateChildButtons()
		if self.settings.showTooltips then
			self.ShowQuantityButton:Enable()
			self.ShowHistoryButton:Enable()
			self.ShowUntrackedQuantityButton:Enable()
		else
			self.ShowQuantityButton:Disable()
			self.ShowHistoryButton:Disable()
			self.ShowUntrackedQuantityButton:Disable()
		end
	end
	
	button = CreateFrame("CHECKBUTTON", "ReagentTrackerOptions_ShowTooltips", self.OptionsFrame, "OptionsSmallCheckButtonTemplate")
	self.ShowTooltipsButton = button
	button:SetPoint("TOPLEFT", self.OptionsFrame, "TOP", -20, 0)
	button:SetSize(26, 26)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Show Tooltips")
	button:SetScript("OnClick", function()
		self.settings.showTooltips = not self.settings.showTooltips
		updateChildButtons()
	end)
	button:SetScript("OnShow", function(button) button:SetChecked(self.settings.showTooltips) updateChildButtons() end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Attach a tooltip with additional information to tracked item tooltips.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	--show quantity info button
	button = CreateFrame("CHECKBUTTON", "ReagentTrackerOptions_ShowQuantity", self.OptionsFrame, "OptionsSmallCheckButtonTemplate")
	self.ShowQuantityButton = button
	button:SetPoint("TOPLEFT", self.ShowTooltipsButton, "BOTTOMLEFT", 26, 2)
	button:SetSize(26, 26)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Show Quantity Info")
	button:SetScript("OnClick", function() self.settings.showQuantityInfo = not self.settings.showQuantityInfo end)
	button:SetScript("OnShow", function(button) button:SetChecked(self.settings.showQuantityInfo) end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Show the character quantity information in tooltips.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	--show history info button
	button = CreateFrame("CHECKBUTTON", "ReagentTrackerOptions_ShowHistory", self.OptionsFrame, "OptionsSmallCheckButtonTemplate")
	self.ShowHistoryButton = button
	button:SetPoint("TOPLEFT", self.ShowQuantityButton, "BOTTOMLEFT", 0, 2)
	button:SetSize(26, 26)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Show History Info")
	button:SetScript("OnClick", function() self.settings.showHistoryInfo = not self.settings.showHistoryInfo end)
	button:SetScript("OnShow", function(button) button:SetChecked(self.settings.showHistoryInfo) end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Show the history information in tooltips.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	--show untracked quantity button
	button = CreateFrame("CHECKBUTTON", "ReagentTrackerOptions_ShowUntrackedQuantity", self.OptionsFrame, "OptionsSmallCheckButtonTemplate")
	self.ShowUntrackedQuantityButton = button
	button:SetPoint("TOPLEFT", self.ShowHistoryButton, "BOTTOMLEFT", 0, 2)
	button:SetSize(26, 26)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Show Untracked Quantities")
	button:SetScript("OnClick", function() self.settings.showUntrackedQuantity = not self.settings.showUntrackedQuantity end)
	button:SetScript("OnShow", function(button) button:SetChecked(self.settings.showUntrackedQuantity) end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Show character's quantities for untracked items.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	 

	 
	--TABS
	--reagent tab
	button = CreateFrame("BUTTON", self:GetName().."Tab1", self, "CharacterFrameTabButtonTemplate")
	self.ReagentTabButton = button
	button:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, 7)
	button:SetText("Reagents")
	button:SetID(1)
	button.page = 1
	button:SetScript("OnClick", function(button)
		self:Tab_OnClick(button)
	end)
		
	--currency tab
	button = CreateFrame("BUTTON", self:GetName().."Tab2", self, "CharacterFrameTabButtonTemplate")
	self.CurrencyTabButton = button
	button:SetPoint("LEFT", self.ReagentTabButton, "RIGHT", -15, 0)
	button:SetText("Currencies")
	button:SetID(2)
	button.page = 2
	button:SetScript("OnClick", function(button)
		self:Tab_OnClick(button)
	end)
		
	--options tab
	button = CreateFrame("BUTTON", self:GetName().."Tab3", self, "CharacterFrameTabButtonTemplate")
	self.OptionsTabButton = button
	button:SetPoint("LEFT", self.CurrencyTabButton, "RIGHT", -15, 0)
	button:SetText("Options")
	button:SetID(3)
	button.page = 3
	button:SetScript("OnClick", function(button)
		self:Tab_OnClick(button)
	end)
	
	self:Hide()
end

--additional tooltips for items
function addon:LoadReagentTooltip()
	--for gametooltips
	self.ReagentTooltip = CreateFrame("GameTooltip", "ReagentTooltip", UIParent, "GameTooltipTemplate")
	self.ReagentTooltip:Hide()
	
	--for inforeftooltips
	self.ReagentTooltip2 = CreateFrame("GameTooltip", "ReagentTooltip2", UIParent, "GameTooltipTemplate")
	self.ReagentTooltip2:Hide()
	
	--for drag frame when reordering reagents
	self.DragFrameTooltip = CreateFrame("GameTooltip", "DragFrameTooltip", UIParent, "GameTooltipTemplate")
	self.DragFrameTooltip:Hide()	
end

--timers for pacing the bag and currency update intervals
function addon:LoadTimers()
	--reagent timer. linked to its own frame so its always watching
	self.bagThrottle = 2
	self.curThrottle = 5
	
	local frame = CreateFrame("FRAME", "ReagentTrackerTimer", UIParent)
	frame:SetScript("OnUpdate", function(frame, elapsed) self:TrackerTimerOnUpdate(elapsed) end)
end

function addon:TrackerTimerOnUpdate(elapsed)
	self.bagThrottle = self.bagThrottle + elapsed
	self.curThrottle = self.curThrottle + elapsed
	
	if self.affectedBags and self.bagThrottle >= 2 then
		self.bagThrottle = 0
		self:ReagentUpdate(self:BuildReagentUpdateList(), true)
	end
	
	if self.currencyUpdate and self.curThrottle >= 5 then
		self.curThrottle = 0
		self.currencyUpdate = false
		if self.TrackerFrame:IsVisible() then
			self:ReagentUpdate({}, false)
		else
			self.currencyUpdateDelayed = true
		end
	end
end

--prepare currency popup and add a button for selecting tracking
function addon:LoadCurrencyTabChanges()
	local button
	TokenFramePopup:SetHeight(132)
	
	button = CreateFrame("CHECKBUTTON", "ReagentTracker_CurrencyButton", TokenFramePopup, "OptionsSmallCheckButtonTemplate")
	self.ReagentTrackerCurrencyButton = button
	button:SetPoint("TOPLEFT", "TokenFramePopupBackpackCheckBox", "BOTTOMLEFT", 0, 4)
	button:SetSize(26, 26)
	button.text = _G[button:GetName().."Text"]
	button.text:SetText("Track Reagent")
	button.text:SetTextColor(1, 1, 1)
	button:SetScript("OnClick", function() self:ToggleCurrencyTracking(TokenFrame.selectedID) end)
	button:SetScript("OnEnter", function(button) self:Tooltip_GameTooltip(button, GameTooltip, "Check to add this currency to the reagent tracker.") end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
end

--update currency button based on tracking
function addon:UpdateCurrencyButton()
	local itemLink = GetCurrencyListLink(TokenFrame.selectedID)
	local checked = self.settings.currencyWatchList[itemLink]
	self.ReagentTrackerCurrencyButton:SetChecked(checked or false)
end

--popup menu for editting reagents
function addon:LoadEditPopup()
	local frame = CreateFrame("FRAME", "ReagentTracker_ReagentPopup")
	frame.displayMode = "MENU"
	frame.initialize = function(frame, level) self:EditPopup_OnLoad() end
	frame:Hide()
end

--create a new container frame for a reagent
function addon:UpdateTrackerFrameCreateFrame(id)
	local frame, fontString, button, texture
	
	frame = CreateFrame("FRAME", nil, self.ReagentTrackerContent)
	self.ReagentDisplay[id] = frame
	frame:SetHeight(20)
	if id == 1 then
		frame:SetPoint("TOPLEFT", self.ReagentTrackerContent, "TOPLEFT", 0, -1)
		frame:SetPoint("TOPRIGHT", self.ReagentTrackerContent, "TOPRIGHT", 0 , -1)
	else
		frame:SetPoint("TOPLEFT", self.ReagentDisplay[id - 1], "BOTTOMLEFT", 0, -2)
		frame:SetPoint("TOPRIGHT", self.ReagentDisplay[id - 1], "BOTTOMRIGHT", 0, -2)
	end
	frame:EnableMouse(true)
	frame:SetID(id)
	
	frame:SetBackdrop({
		bgFile = "Interface\\CHATFRAME\\CHATFRAMEBACKGROUND",
		edgeFile="Interface\\CHATFRAME\\CHATFRAMEBACKGROUND",
		tile=1, tileSize=10, edgeSize=1, 
		insets={left=1, right=1, top=1, bottom=1}
	})
	frame:SetBackdropColor(0, 0, 0, 0)
	frame:SetBackdropBorderColor(0, 0, 0, 1)
	
	frame:SetScript("OnLeave", function() GameTooltip:Hide() end)
	frame:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
	frame:RegisterForDrag("RightButton")
	
	--bg texture
	texture = frame:CreateTexture()
	frame.bgColor = texture
	texture:SetAllPoints(frame)
	texture:SetHeight(frame:GetHeight())
	texture:SetTexture(1, 1, 1, 0.6)
	texture:SetGradient("HORIZONTAL", 0, 0, 0, 0, 0, 0)
	
	--color band
	texture = frame:CreateTexture()
	frame.colorBand = texture
	texture:SetPoint("TOPLEFT", frame)
	texture:SetSize(3, frame:GetHeight())
	texture:SetTexture(0, 0, 0, 1)
	
	--reagent	
	fontString = frame:CreateFontString()
	frame.reagentName = fontString
	fontString:SetPoint("LEFT", frame, "LEFT", 6, 0)
	fontString:SetFontObject("GameFontNormal")
	
	--trend
	button = CreateFrame("FRAME", nil, frame)
	frame.trendFrame = button
	button:SetPoint("RIGHT", frame, "RIGHT", -4, 0)
	button:SetWidth(50)
	button:SetHeight(20)
	
	button:SetScript("OnEnter", function(button) self:ShowAdvancedInfoTooltip(button) end)
	button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	button:SetScript("OnMouseUp", function() self:GetCursorReagent() end)
	button:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
	
	fontString = frame.trendFrame:CreateFontString()
	frame.trend = fontString
	fontString:SetPoint("RIGHT", frame.trendFrame, "RIGHT", 0, 0)
	fontString:SetFontObject("GameFontNormal")
	fontString:SetTextColor(1, 1, 1, 1)
	fontString:SetWidth(50)
	fontString:SetJustifyH("RIGHT")
	
	--quantity
	button = CreateFrame("BUTTON", nil, frame)
	frame.quantity = button
	button:SetPoint("RIGHT", frame.trendFrame, "LEFT", -4, 0)
	button:SetNormalFontObject("GameFontNormal")
	button:SetWidth(60)
	button:SetHeight(20)
	button:SetText("test")
	button:SetPushedTextOffset(0, 0)
	
	button:SetScript("OnClick", function() self:GetCursorReagent() end)
	button:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
end

--build display table for reagent. update frame id if provided
function addon:CalculateReagentDisplayInfo(reagent, id, watchList)
	--store values here
	local info = {
		reagent = reagent,
		isGold = false,
		--soulbound
		--trend
		--totalCount
		--displayCount
		--individualCount
		--itemLink
		--displayName
		--average
		--sum
		--historyInfo
	}
	
	--no overriding watchlist supplied?
	if not watchList then
		watchList = self.visibleWatchList
	end
	
	--soulbound?
	info.soulbound = watchList[info.reagent].soulbound
	
	--currencies
	if self.tabDisplay == 2 then 
		info.soulbound = true
	end
	
	--gold
	if info.reagent == "gold" then
		info.soulbound = false
		info.isGold = true
	end

	--trend
	local cachedTrend = self.localReagents[info.reagent].cachedTrend
	if info.soulbound then
		cachedTrend = self.localReagents[info.reagent][UnitName("player")].cachedTrend
	end
	info.trend = cachedTrend or 0

	--total and individual counts
	info.totalCount, info.individualCount = self:AllCharReagentCountCurrent(info.reagent)
	
	--displayed count
	info.displayCount = info.totalCount
	if info.soulbound then
		info.displayCount = info.individualCount[UnitName("player")]
	end
	
	--item link
	if self.tabDisplay == 1 then
		info.itemLink = self:SaveReagentLinkInfo(info.reagent)
		
		--convert to an itemID if possible
		if not tonumber(info.reagent) then
			info.reagent = self:UpdateReagentName(info.itemLink) or reagent
		end
	else
		info.itemLink = info.reagent
	end
	info.displayName = info.itemLink or info.reagent
	
	--stuff for gold currency
	if info.isGold then
		info.displayName = "|cFFFFD100Gold"
		info.displayCount = self:FormatGoldString(info.displayCount, true)
	end
			
	if id then
		self:PopulateReagentFrame(id, info)
	else
		return info
	end
end

--take calculated values and plug them into the frame
function addon:PopulateReagentFrame(id, info)
	local frame = self.ReagentDisplay[id]
	local trendButton = frame.trendFrame
	local quantityButton = frame.quantity
	
	frame.reagent = info.reagent
		
	--text display
	frame.reagentName:SetText("|cFFFFFFFF"..info.displayName)
	frame.quantity:SetText(info.displayCount)
	frame.trend:SetText(self:ColorCodeString(info.trend, self.prefix.trend, info.isGold, info.isGold))
	
	--tooltips
	--build count list
	local quantityInfo = {}
	for character, count in pairs(info.individualCount) do
		if count > 0 then
			local countColor = {r =1, g= 1, b = 1}
			if character == UnitName("player") then
				countColor = {r =0, g= 1, b = 0}
			end
			local tooltipInfo = {
				text = character, 
				color = countColor,
				textR = info.isGold and self:FormatGoldString(count).."   " or count,
				colorR = countColor,
			}
			table.insert(quantityInfo, tooltipInfo)
		end
	end
	table.sort(quantityInfo, function(a, b) return a.text < b.text end)	
	table.insert(quantityInfo, 1, {text = "Character", textR = "Quantity"})
	
	--show a total for gold with silver and copper
	if info.isGold then
		table.insert(quantityInfo, {text = " "})
		table.insert(quantityInfo, {text = "Total", textR = self:FormatGoldString(info.totalCount).."   "})
	end
	
	--color bg color for soulbound items
	if self.tabDisplay == 1 and info.soulbound then
		--purple
		frame.bgColor:SetGradient("HORIZONTAL", 0.07, 0, 0.1, 0, 0, 0)
		frame:SetBackdropBorderColor(0.07, 0, 0.1, 1)
	else
		--black
		frame.bgColor:SetGradient("HORIZONTAL", 0, 0, 0, 0, 0, 0)
		frame:SetBackdropBorderColor(0, 0, 0, 1)
	end
		
	--find clicked frames id
	frame:SetScript("OnMouseDown", function()
		self.clickedFrameID = frame:GetID()
	end)
	
	--right click menu
	if info.isGold then
		frame:SetScript("OnMouseUp", function() self:GetCursorReagent() end)	--always check for item drops
	else
		frame:SetScript("OnMouseUp", function(_, mouseButton)
			if CursorHasItem() then
				self:GetCursorReagent()
			elseif mouseButton == "RightButton" then
				--menu
				Lib_ToggleDropDownMenu(1, nil, ReagentTracker_ReagentPopup, "cursor")
			else
				--link
				if info.itemLink then
					if IsModifiedClick("CHATLINK") then
						--paste chat link
						ChatEdit_InsertLink(info.itemLink)
					else
						--tooltip with info
						GameTooltip:SetOwner(frame, "ANCHOR_LEFT", 0, -20)
						GameTooltip:SetHyperlink(info.itemLink)
						GameTooltip:Show()
					end
				end
			end
		end)
	end

	--quantity tooltip
	if #quantityInfo > 1 then
		frame.quantity:SetScript("OnEnter", function(quantityButton) self:Tooltip_GameTooltip(quantityButton, GameTooltip, quantityInfo) end)
	else
		frame.quantity:SetScript("OnEnter", nil)
	end
	frame.quantity:SetScript("OnLeave", function() GameTooltip:Hide() end)
	
	--order changing
	frame:SetScript("OnDragStart", function(frame, mouseButton)
		if mouseButton == "RightButton" then
			self:FrameDragStart(frame)
		end
	end)
	
	--frame is good, show it
	frame:Show()
end

--build and show complimentary info tooltip for tracked reagents
function addon:ShowReagentTooltip(frame)
	if self.settings.showTooltips then
		local tooltip = self.ReagentTooltip
		if frame:GetName() == "ItemRefTooltip" then
			tooltip = self.ReagentTooltip2
		end
		
		local name, link = frame:GetItem()
		
		tooltip:SetOwner(frame, "ANCHOR_NONE")
		tooltip:ClearAllPoints()
		tooltip:SetPoint("TOPLEFT", frame, "BOTTOMLEFT")
		tooltip:SetScale(0.8)
		
		if link then
			itemID = tonumber(link:match("|Hitem:(%d+)"))
			
			if self.settings.watchList[itemID] then
				local info = self:CalculateReagentDisplayInfo(itemID, nil, self.settings.watchList)
				
				--total count
				tooltip:AddDoubleLine("Total", "|cFFFFFFFF"..info.totalCount.."|r")
				
				if self.settings.showQuantityInfo then
					--space
					tooltip:AddLine(" ")
					
					--character counts --TODO sorting
					tooltip:AddDoubleLine("Character", "Quantity")
					for character, charCount in pairs(info.individualCount) do
						local countColor = "|cFFFFFFFF"
						if character == UnitName("player") then
							countColor = "|cFF00FF00"
						end
						if charCount > 0 then
							tooltip:AddDoubleLine(countColor..character.."|r", countColor..charCount.."|r")
						end
					end
				end
				
				if self.settings.showHistoryInfo then
					local advancedInfo = self:CalculateAdvancedInfo(itemID, self.settings.watchList)
					
					if advancedInfo then
						--space
						tooltip:AddLine(" ")
						
						--history info
						tooltip:AddLine("History")
						tooltip:AddLine(advancedInfo.historyString)
						tooltip:AddLine(" ")
						tooltip:AddLine("Two week stats:")
						tooltip:AddLine("|cFFFFFFFFTrend:|r "..self:ColorCodeString(advancedInfo.trend, self.prefix.trend, advancedInfo.isGold, advancedInfo.isGold))
						tooltip:AddLine("|cFFFFFFFFSum:|r "..self:ColorCodeString(advancedInfo.sum, nil, nil, advancedInfo.isGold))
						tooltip:AddLine("|cFFFFFFFFAverage per day:|r "..self:ColorCodeString(advancedInfo.average, nil, nil, advancedInfo.isGold))
					end
				end
				
				tooltip:Show()
			elseif self.settings.showUntrackedQuantity then
				local totalCount = GetItemCount(itemID, true, true)
				local bagCount = GetItemCount(itemID, false, true)
				local bankCount = totalCount - bagCount
				if totalCount > 0 then
					tooltip:AddDoubleLine("Character Total", "|cFFFFFFFF"..totalCount.."|r")
					
					if bagCount > 0 then
						tooltip:AddDoubleLine("Backpack", "|cFFFFFFFF"..bagCount.."|r")
					end
					
					if bankCount > 0 then
						tooltip:AddDoubleLine("Bank", "|cFFFFFFFF"..bankCount.."|r")
					end
				end
				tooltip:Show()
			end
		end
	end
end

--take a list of numbers and convert to a list of differences between the values
function addon:ListDelta(list)
	local deltaList = {}
	if #list > 1 then
		for id = 1, #list - 1 do
			local delta = list[id + 1] - list[id] 
			table.insert(deltaList, delta)
		end
	else
		deltaList = list
	end
	
	return deltaList
end

-- update and display gametooltip with list of info
--[[	- frame is the widget to attach to
		- tooltip is the tooltip frame being used (GameTooltip or ItemRefTooltip or something else)
		- info = {
			{
				text = "",							--required
				color = {r=1, g=1, b=1},
				textR = "",
				colorR = {r=1, g=1, b=1},
				wrap
			}
		}
				-=OR=-
		- info = "text"
]]
function addon:Tooltip_GameTooltip(frame, tooltip, info)
	if info then
		tooltip:SetOwner(frame, "ANCHOR_RIGHT")
		if type(info) == "table" then
			for id = 1, #info do
				info[id].color = info[id].color or {}
				if info[id].textR then
					info[id].colorR = info[id].colorR or {}
					tooltip:AddDoubleLine(info[id].text, info[id].textR, info[id].color.r, info[id].color.g, info[id].color.b, info[id].colorR.r, info[id].colorR.g, info[id].colorR.b, info[id].wrap)
				else
					tooltip:AddLine(info[id].text, info[id].color.r, info[id].color.g, info[id].color.b, info[id].wrap)
				end
			end
		else
			tooltip:AddLine(info)
		end
		
		tooltip:Show()
	end
end

--handle tab buttons on the bottom of frame
function addon:Tab_OnClick(button)
	local id = button:GetID()
	--tab stuff
	PanelTemplates_SetTab(self, id)
	self.tabDisplay = id
	local offset
	--change title
	if id ==  1 then
		self.TrackerFrame:Show()
		self.ReagentTrackerReagentTitle:SetText("Reagent")
		self.visibleWatchList = self.settings.watchList
		self.visibleSortedList = self.settings.sortedReagentList
		
		self.ReagentTrackerTrendTitle:Show()
		self.ReagentTrackerQuantityTitle:Show()
		
		self:UpdateTrackerFrame()
		self.ReagentTrackerWindow:SetVerticalScroll(self.prevReagentOffset or 0)
	elseif id == 2 then
		self.TrackerFrame:Show()
		self.ReagentTrackerReagentTitle:SetText("Currency")
		self.visibleWatchList = self.settings.currencyWatchList
		self.visibleSortedList = self.settings.sortedCurrencyList
		
		self.ReagentTrackerTrendTitle:Show()
		self.ReagentTrackerQuantityTitle:Show()
		
		self:UpdateTrackerFrame()
		self.ReagentTrackerWindow:SetVerticalScroll(self.prevCurrencyOffset or 0)
	else
		self.OptionsFrame:Show()		--hiding of scrollframe is handled by onshow handler
		self.ReagentTrackerReagentTitle:SetText("Add Item")
		
		self.ReagentTrackerTrendTitle:Hide()
		self.ReagentTrackerQuantityTitle:Hide()
	end

	PlaySound("igCharacterInfoTab")
end

function addon:EditPopup_OnLoad()
	local info
	
	local id = self.clickedFrameID
	local reagent = self.ReagentDisplay[id].reagent
		
	if self.tabDisplay == 1 then
		--soulbound
		if reagent then
			info = Lib_UIDropDownMenu_CreateInfo()
			info.text = "Soulbound"
			info.isNotRadio = true
			info.checked = self.visibleWatchList[reagent].soulbound
			info.func = function()
				self.visibleWatchList[reagent].soulbound = not self.visibleWatchList[reagent].soulbound
				self:CalculateReagentDisplayInfo(reagent, id)
			end
			Lib_UIDropDownMenu_AddButton(info)
			
			--space
			info = Lib_UIDropDownMenu_CreateInfo()
			info.isTitle = truec
			info.notCheckable = 1
			Lib_UIDropDownMenu_AddButton(info)
		end
	end

	--delete
	info = Lib_UIDropDownMenu_CreateInfo()
	info.text = "Stop Tracking"
	info.func = function()
		if self.tabDisplay == 1 then
			self:RemoveReagent(reagent)
		else
			self:RemoveCurrency(reagent)
		end
	end
	info.notCheckable = 1
	Lib_UIDropDownMenu_AddButton(info)
end

--get id of grabbed reagent and make frames watch for dragging end
function addon:FrameDragStart(frame)
	if frame.reagent then
		self.draggedReagentID = frame.reagent
		self:FrameDragFrameSetup()
		frame:SetScript("OnDragStop", function() self.DragFrame:Hide() end)
		
		--populate and show tooltip for dragged reagent
		self.DragFrame:Show()
		
		self.DragFrameTooltip:SetOwner(UIParent, "ANCHOR_NONE")
		self.DragFrameTooltip:ClearAllPoints()
		self.DragFrameTooltip:SetPoint("TOPLEFT", self.DragFrame)
		self.DragFrameTooltip:SetScale(0.8)
		
		local link = frame.reagent	--currencies covered here
		if self.tabDisplay == 1 then
			link = self:SaveReagentLinkInfo(frame.reagent) or frame.reagent
		else
			if frame.reagent == "gold" then
				link = "|cFFFFD100Gold"
			end
		end
		self.DragFrameTooltip:AddLine(link)
		
		self.DragFrameTooltip:Show()
	end
end

--apply dragged frame info to new location. clean up display and update
function addon:FrameDragEnd(frame)
	self:FrameDragFrameCleanup()
	local oldReagent = self.draggedReagentID
	local newReagent = frame.reagent
	
	local newLoc
	for id, reagent in pairs(self.visibleSortedList) do
		if reagent == newReagent then
			newLoc = id
		end
	end
	
	if newLoc then
		tDeleteItem(self.visibleSortedList, oldReagent)
		table.insert(self.visibleSortedList, newLoc, oldReagent)
		self:UpdateTrackerFrame()
	end
	self.draggedReagentID = nil
end

--prepare tracker frames for receiving drags
function addon:FrameDragFrameSetup()
	local id = 1
	while self.ReagentDisplay[id] ~= nil do
		self.ReagentDisplay[id].trendFrame:EnableMouse(false)
		self.ReagentDisplay[id].quantity:EnableMouse(false)
		self.ReagentDisplay[id]:SetScript("OnReceiveDrag", function(frame) self:FrameDragEnd(frame) end)
		id = id + 1
	end
end

--stop tracker frames from receiving drags
function addon:FrameDragFrameCleanup()
	local id = 1
	while self.ReagentDisplay[id] ~= nil do
		self.ReagentDisplay[id].trendFrame:EnableMouse(true)
		self.ReagentDisplay[id].quantity:EnableMouse(true)
		self.ReagentDisplay[id]:SetScript("OnReceiveDrag", function() self:GetCursorReagent() end)
		self.ReagentDisplay[id]:SetScript("OnDragStop", nil)
		id = id + 1
	end
end

--return a currency string with coin icons and round it to gold only if necessary
function addon:FormatGoldString(value, round)
	value = math.abs(value)
	
	if round then
		value = math.floor(value / 10000) * 10000		--get rid of silver and copper
	end

	if value ~= 0 then
		return GetCoinTextureString(value)
	else
		return "0"
	end
end

--format text string based on color table provided and whether it is gold or not
function addon:ColorCodeString(textString, colorTable, goldCoins, goldOnly)
	if not colorTable then
		colorTable = self.prefix.reg
	end
	
	--color the pattern found with appropriate color and formatting
	local function recolorFunc(number)
		local comp = tonumber(number)
		--get rid of silver and copper
		if goldOnly then
			number = math.floor(number / 10000)
		end
		--add coin textures?
		if goldCoins then
			--compensate for gold only division
			if goldOnly then
				number = number * 10000
			end
			number = self:FormatGoldString(number)
		end
		if comp < 0 then
			return colorTable.less..number.."|r" 		--red
		elseif comp > 0 then
			return colorTable.more..number.."|r"		--green
		else
			return colorTable.equal..number.."|r"		--blue
		end	
	end
	
	--search for numbers and format them
	local newString = string.gsub(textString, "[%-]?[0-9]+[%.]*[0-9]*", recolorFunc)		--optional "-", number, optional decimal and fraction

	return newString
end

--setup default values for frame
function addon:SetFrameDefaults(frame)
	frame.reagentName:SetText("...")
	frame.quantity:SetText("")
	frame.trend:SetText("")
end

--point reagent table at server and faction table for current character
function addon:InitializeServerFactionInfo()
	local player = self.player.name
	local faction = self.player.faction
	local server = self.player.server
	
	if not self.reagents[server] then
		self.reagents[server] = {}
	end
	
	if not self.reagents[server][faction] then
		self.reagents[server][faction] = {}
	end
		
	self.localReagents = self.reagents[server][faction]
end

--move from self.reagents[reagent] to self.reagents[server][faction][reagent] to separate unrelated reagent info
function addon:UpdateReagentFormat()
	self.settings.formatUpdate = true
	local reagentTable = copyTable(self.reagents, true, true)
	
	local player = self.player.name
	local faction = self.player.faction
	local server = self.player.server
	
	wipe(self.reagents)
	self.reagents[server] = {}
	self.reagents[server][faction] = copyTable(reagentTable, true, true)
end

--build and return a list containing any reagents found in a bag that is being tracked
function addon:BuildReagentUpdateList()
	local updateList = {}

	if self.affectedBags then
		for bagID, slots in pairs(self.affectedBags) do
			if #slots > 0 then
				--slotIDs provided
				for _, slotID in pairs(slots) do
					local itemID = GetContainerItemID(bagID, slotID)
					if self.settings.watchList[itemID] then
						updateList[itemID] = true
					end
				end
			else
				--slotIDs unknown
				for slotID = 1, GetContainerNumSlots(bagID) do
					local itemID = GetContainerItemID(bagID, slotID)
					if self.settings.watchList[itemID] then
						updateList[itemID] = true
					end
				end
			end
		end
		
		self.affectedBags = nil
	end

	return updateList
end

--get the item on the cursor and attempt to add it
function addon:GetCursorReagent()
	local type, data = GetCursorInfo()
	
	if type == "item" then
		self:AddReagent(data)
	end
	
	ClearCursor()
end

--update the reagent list based off scrollframe offset
function addon:UpdateTrackerFrame()
	local frameCount = 6
	FauxScrollFrame_Update(self.ReagentTrackerWindow, #self.visibleSortedList, frameCount, 20, nil, nil, nil, nil, nil, nil, true)
	local offset = FauxScrollFrame_GetOffset(self.ReagentTrackerWindow)
	
	for id = 1, frameCount do
		local reagentID = id + offset
		local reagent = self.visibleSortedList[reagentID]
		local frame = self.ReagentDisplay[id]
			
		if reagent then
			self:CalculateReagentDisplayInfo(reagent, id)		
		else
			self.ReagentDisplay[id]:Hide()
		end
	end	
end

--do and return more stressful calculations for a reagent
function addon:CalculateAdvancedInfo(reagent, watchList)
	if reagent then
		local info = {}
		if not watchList then watchList = self.visibleWatchList end
		local soulbound = watchList[reagent].soulbound or watchList["gold"]
		local character = UnitName("player")
		
		if reagent == "gold" then
			info.isGold = true
			soulbound = false
		end
		
		local periodInfo = self:GetTimePeriodInfo(reagent, soulbound)
		
		--trend
		info.trend = ( math.floor(periodInfo.trend * 100) / 100 ) or self.localReagents[reagent].cachedTrend or 0
		
		--save either to character or reagent depending on soulbound flag
		if soulbound then
			self.localReagents[reagent][character].cachedTrend = info.trend
			self.localReagents[reagent].cachedTrend = nil
		else
			self.localReagents[reagent].cachedTrend = info.trend
			if self.localReagents[reagent][character].cachedTrend then
				self.localReagents[reagent][character].cachedTrend = nil
			end
		end
		
		--delta list
		local historyDelta = self:ListDelta(periodInfo.history)
		
		--sum
		info.sum = 0
		for _, deltaValue in pairs(historyDelta) do
			info.sum = info.sum + deltaValue
		end
		
		--average
		info.average = math.floor((info.sum / #historyDelta) * 100) / 100
		
		--history string
		info.historyString = self:ColorCodeString(table.concat(historyDelta, " "), nil, nil, info.isGold)
		
		return info
	end
end

--show history info for a reagent
function addon:ShowAdvancedInfoTooltip(button)
	local frame = button:GetParent()
	local reagent = frame.reagent
	 
	if reagent then
		local info = addon:CalculateAdvancedInfo(reagent)
		local historyInfo = {
		[1] = {text = "History"},
		[2] = {text = info.historyString, wrap = false},
		[3] = {text = " "},
		[4] = {text = "Two week stats:"},
		[5] = {text = "|cFFFFFFFFSum:|r "..self:ColorCodeString(info.sum, nil, nil, info.isGold)},
		[6] = {text = "|cFFFFFFFFAverage per day:|r "..self:ColorCodeString(info.average, nil, nil, info.isGold)},
	}
		self:Tooltip_GameTooltip(button, GameTooltip, historyInfo)
		frame.trend:SetText(self:ColorCodeString(info.trend, self.prefix.trend, info.isGold, info.isGold))
	end
end

--remove unused textures being saved
function addon:RemoveSavedTextures()
	self.settings.textureCleared = true
	
	for reagent, info in pairs(self.settings.watchList) do
		info.texture = nil
	end
	
	for currency, info in pairs(self.settings.currencyWatchList) do
		info.texture = nil
	end
end

--return date to the day in seconds. pos offset is forward in time neg is backward
function addon:GetDateStamp(timeStamp, offset)
	if not timeStamp then timeStamp = time() end
	local date = date("!*t", timeStamp)
	
	local dayFloor = (date.hour * 3600) + (date.min * 60) + date.sec		--time to be removed
	local dayLen = 24 * 3600		--day in seconds
		
	return timeStamp - dayFloor + ( (offset or 0) * dayLen)
end

--convert old date stamp to new one
function addon:ConvertDateStamps()
	--build table of datestamps with their corresponding number value 
	local timePeriod = {}
	for offset = 0, 15 do
		local time = self:GetDateStamp(nil, -offset)
		local date = date("!*t", time)
		local dateStamp = date.day.."."..date.month.."."..date.year

		timePeriod[dateStamp] = time
	end
	
	--iterate over every reagent and convert
	for server, serverInfo in pairs(self.reagents) do
		for faction, factionInfo in pairs(serverInfo) do
			for reagent, reagentInfo in pairs(factionInfo) do
				for char, info in pairs(reagentInfo) do
					--temp table
					local newTable = {}
					
					for dateString, dateInfo in pairs(info.data) do
						if type(dateString) == "string" then
							if timePeriod[dateString] then
								--populate temp table with old values but new keys
								newTable[timePeriod[dateString]] = copyTable(info.data[dateString], true, true)
								info.data[dateString] = nil
							end
						end
					end
					
					--place table into old location
					info.data = newTable
				end
			end
		end
	end
		
	self.settings.newDateSystem = true	
end

--convert saved table to single values
function addon:ConvertItemCounts()	
	for server, serverInfo in pairs(self.reagents) do
		for faction, factionInfo in pairs(serverInfo) do
			for reagent, reagentInfo in pairs(factionInfo) do
				for char, info in pairs(reagentInfo) do
					for dateString, count in pairs(info.data) do
						if type(count) == "table" then
							info.data[dateString] = count.itemCount
						end
					end
				end
			end
		end
	end
	self.settings.newCountSystem = true	
end

--delete any reagent data for a specific character
function addon:DeleteCharacterData(deletedChar)
	if deletedChar then
		--delete data
		for reagent, reagentInfo in pairs(self.localReagents) do
			for character in pairs(reagentInfo) do
				if deletedChar == character then
					reagentInfo[character] = nil
				end
			end
		end
		
		--reset display of character list
		Lib_UIDropDownMenu_SetSelectedValue(self.CharDropdown, nil)
		Lib_UIDropDownMenu_SetText(self.CharDropdown, "Characters")
		self.DeleteCharButton:Disable()
		
		--refresh data
		self:ReagentUpdate()
		print("All data for "..deletedChar.." has been removed.")
	end
end

--queue up the bagID if it isnt yet
function addon:AddBagIDCheck(bagID, slotID)
	self.affectedBags = self.affectedBags or {}
	self.affectedBags[bagID] = self.affectedBags[bagID] or {}
	
	if slotID then
		if not tContains(self.affectedBags[bagID], slotID) then
			table.insert(self.affectedBags[bagID], slotID)
		end
	end
end

--build list of watched items found in mail attachments. return nil if nothing found
function addon:BuildMailItemList(updateList)
	for slotID = 1, ATTACHMENTS_MAX_SEND do
		local link = GetSendMailItemLink(slotID)
		if link then
			itemID = tonumber(link:match("|Hitem:(%d+)"))
			if self.settings.watchList[itemID] then
				updateList[itemID] = true
			end
		end
	end
end

--update position and display of dragframe anchor
function addon:UpdateDragFrame(frame)
	local x, y = GetCursorPosition()
	x, y = x / UIParent:GetEffectiveScale(), y / UIParent:GetEffectiveScale()
	frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
	
	SetCursor("ITEM_CURSOR")
	
	if not self.draggedReagentID then
		frame:Hide()
	end
end

--initialize and populate character dropdown
function addon:CharacterListDropdownInitialization(dropDown, level)
	for character, charInfo in pairs(self.localReagents["gold"]) do
		if type(charInfo) == "table" then
			local function onClickFunc()
				Lib_UIDropDownMenu_SetText(dropDown, character)
				Lib_UIDropDownMenu_SetSelectedValue(dropDown, character)
				self.DeleteCharButton:Enable()
			end
			
			info = Lib_UIDropDownMenu_CreateInfo()
			info.text = character
			info.value = character
			info.func = onClickFunc
			Lib_UIDropDownMenu_AddButton(info, level)
		end
	end
end






















